module Orator

  # A base for the orators to split off of.  It handles contexts and running
  # methods.  Most of the methods defined in this  class are prefixed with
  # two dashes to prevent polluting the namespace.
  #
  # @abstract
  class Base

    # Initialize the orator with the context.
    #
    # @param context [Object]
    def initialize(context)
      @__context = context
    end

    # This triggers an event if it's defined on this orator.  If it is not
    # defined, it raises an error.
    #
    # @param event [Symbol, String] the event to trigger.
    # @raise [NoMethodError] if the event isn't defined on the orator.
    # @return [Array<Object>, nil] the values of the trigger, or nil if the
    #   event was prevented.
    def __trigger(event, *args)
      orator = event[0..event.index('.')-1]
      method_name = event[event.index('.')+1..-1]
      __check_event(orator, method_name)

      @event = [orator, method_name]
      __run_callbacks(:before, args)
      out = __run_event(method_name, args) unless @__prevent_event
      __run_callbacks(:after, args)

      out
    end

    # Returns the current event that the orator is handling in `orator.event`
    # form.
    #
    # @return [String]
    def full_event
      @event.join('.')
    end

    # Returns the current event that the orator is handling in `event` form.
    #
    # @return [String]
    def event
      @event[1]
    end

    # Checks the event list to see if this class responds to it.
    #
    # @return [Set<Symbol, Proc>, nil]
    def __method_exists?(method)
      __event_list[method]
    end

    # A map of the events that determine how this class will respond.
    #
    # @return [Hash]
    # @see [::event_list]
    def __event_list
      self.class.event_list
    end

    # This tells the orator to not execute the events in the class.  This
    # should only be called in a `before` block.
    def prevent_event
      @__prevent_event = true
    end

    # Handles missing methods by delegating them to the context.  Should
    # never be called directly.
    #
    # @return [Object]
    def method_missing(method, *args, &block)
      super unless respond_to_missing?(method)

      @__context.public_send(method, *args, &block)
    end

    # Lets Ruby know that there are some undefined methods on this class.
    #
    # @return [Boolean]
    def respond_to_missing?(method, include_private = false)
      @__context.respond_to?(method, include_private)
    end

    # This forwards {#send} on to the context.
    #
    # @return [Object]
    def send(*args, &block)
      @__context.send(*args, &block)
    end

    private

    # Run the event.
    def __run_event(method_name, args)
      result = __event_list[method_name].map do |responder|
        responder = method(responder) if responder.is_a? Symbol
        instance_exec *args, &responder
      end

      result
    end

    # Check the event, making sure that this orator can respond to it.
    def __check_event(orator, method)
      if orator != self.class.orator_name
        raise NoMethodError,
          "Event #{orator} does not match #{self.class.name}"
        return
      elsif !__method_exists?(method)
        raise NoMethodError,
          "Method #{method} does not exist on #{self.class.name}"
        return
      end
    end

    # Runs the callbacks.
    def __run_callbacks(type, args)
      self.class.send("#{type}_list").map do |cb|
        if cb.is_a? Symbol
          method(cb).call(*args)
        else
          instance_exec(*args, &cb)
        end
      end
    end

    module ClassMethods

      # This manages the orator's name.  If an argument is passed, the name
      # is set to that.  Otherwise, it returns the orator's name.
      #
      # @param value [String, nil] the new name.
      # @return [String] the orator name.
      def orator_name(value = nil)
        if value
          @_orator_name = value
        end

        @_orator_name
      end

      # This registers a method for this orator.
      #
      # @param event [Symbol] the event to register it to.
      # @return [void]
      def on(event, method = nil, &block)
        if event.is_a? Hash
          method = event.values.first
          event  = event.keys.first
        end
        
        event_list[event.to_s] ||= []
        
        if block_given?
          event_list[event.to_s] << block       
        else
          event_list[event.to_s] << (method || event).to_sym
        end
      end

      # The event matchings.  The keys are the event name, the values can be
      # an array of mappings or a the mapping itself.
      #
      # @return [Hash] the event list.
      def event_list
        @event_list ||= {}
      end

      # This registers this orator with the event handler by telling it what
      # methods it responds to.
      #
      # @param event_handler [EventHandler]
      def register_with(event_handler)
        event_list.keys.each do |event|
          event_handler.on("#{self.orator_name}.#{event}", self)
        end
      end

      # This handles the before callbacks that our orator needs.  It can accept
      # a method or a block, and given the two, chooses the method over the
      # block.
      #
      # @example
      #   before do; something; end
      # @example
      #   before :some_method
      # @param method [Symbol] the method that will be called for the before
      #   callback.
      # @yields []
      def before(method = nil, &block)
        before_list << (method or block)
      end

      # This sets up the before list on the first call, and returns the list
      # on the first and subsequent calls.
      #
      # @return [Set<Symbol, Block>] the before callback list.
      def before_list
        @before_list ||= Set.new
      end

      # This handles the after callbacks that our operator needs.  It can accept
      # a method or a block, and given the two, chooses the method over the
      # block.
      #
      # @example
      #   after do; something; end
      # @example
      #   after :some_method
      # @param method [Symbol] the method that will be called for the after
      #   callback.
      # @yields []
      def after(method = nil, &block)
        after_list << (method or block)
      end

      # This sets up the after list on the first call, and returns the list on
      # the first and subsequent calls.
      #
      # @return [Set<Symbol, Block>] the after callback list.
      def after_list
        @after_list ||= Set.new
      end

      # This is called when this class is inherited from.  We use this to make
      # sure that each subclass inherits the event list as well as the before
      # and after methods.
      #
      # @param klass [Base]
      def inherited(klass)
        klass.instance_variable_set(:@event_list,  event_list.dup )
        klass.instance_variable_set(:@before_list, before_list.dup)
        klass.instance_variable_set(:@after_list,  after_list.dup )
      end
    end

    extend ClassMethods
    orator_name ''

  end

end
